// Copyright (C) 2000 Nicholas Schmitt nick@math.umass.edu

#ifndef MathReal_INCLUDED
#define MathReal_INCLUDED

#include <iostream.h> 
#include <math.h>
#include <float.h>
#include "System.H"
#include "Use.h"

#undef isnan

namespace math {

class MathReal
{

public:

  // pi
  static double const pi;

  // return the minimum
  static inline float min(float, float);
  static inline double min(double, double);
  static inline double long min(double long, double long);

  // return the maximum
  static inline float max(float, float);
  static inline double max(double, double);
  static inline double long max(double long, double long);
  
  // return the absolute value
  static inline float abs(float);
  static inline double abs(double);
  static inline double long abs(double long);

  // factorial
  static double factorial( int unsigned n );

  // chop: 0 if less than tolerance
  static inline float chop(
    float x, float tolerance = 1.0e-6 );
  static inline double chop(
    double x, double tolerance = 1.0e-10 );
  static inline double long chop(
    double long x, double long tolerance = 1.0e-30 );

  // return true iff is a nan
  static inline bool isnan(float);
  static inline bool isnan(double);
  static inline bool isnan(double long);

  // effectively_zero
  static inline bool effectively_zero(
    float x, float precision = 100.0 * FLT_MIN );
  static inline bool effectively_zero(
    double x, double precision = 100.0 * DBL_MIN );
  static inline bool effectively_zero(
    double long x, double long precisoin = 100.0 * LDBL_MIN );

  /// print in c format
  static void print( ostream &o, float );
  static void print( ostream &o, double );
  static void print( ostream &o, double long );

  /// print in c format; append a '\n'
  static inline void println( ostream &o, float );
  static inline void println( ostream &o, double );
  static inline void println( ostream &o, double long );

  /// print in mathematica format
  static void printM( ostream &o, float );
  static void printM( ostream &o, double );
  static void printM( ostream &o, double long );

  /// print in mathematica format; append a '\n'
  static inline void printMln( ostream &o, float );
  static inline void printMln( ostream &o, double );
  static inline void printMln( ostream &o, double long );

protected:

private:

};


float
MathReal::
min( float x, float y )
{
  return x <= y ? x : y;
}

double
MathReal::
min( double x, double y )
{
  return x <= y ? x : y;
}

double long
MathReal::
min( double long x, double long y )
{
  return x <= y ? x : y;
}


float
MathReal::
max( float x, float y )
{
  return x >= y ? x : y;
}

double
MathReal::
max( double x, double y )
{
  return x >= y ? x : y;
}

double long
MathReal::
max( double long x, double long y )
{
  return x >= y ? x : y;
}

float
MathReal::
abs( float x )
{
  return ::fabsf(x);
}

double
MathReal::
abs( double x )
{
  return ::fabs(x);
}

double long
MathReal::
abs( double long x )
{
#if defined(__sgi) && !defined(__GNUG__) && defined(_COMPILER_VERSION)
#if _COMPILER_VERSION >= 730
  return ::fabsl(x);
#else
  return ::absl(x);
#endif
#else
  return ::fabsl(x);
#endif
}

bool
MathReal::
isnan( float x )
{
  // ???
  return ::isnan(x);
}

bool
MathReal::
isnan( double x )
{
  return ::isnan(x);
}

bool
MathReal::
isnan( double long x )
{
#if defined(__sgi) && defined(_COMPILER_VERSION)
#if _COMPILER_VERSION >= 730
  return ::isnanl(x);
#else
  USE(x);
  DIE();
  return true;
#endif
#else
  return ::isnanl(x);
#endif
}

bool
MathReal::
effectively_zero( float x, float precision )
{
  return abs(x) < precision;
}

bool
MathReal::
effectively_zero( double x, double precision )
{
  return abs(x) < precision;
}

bool
MathReal::
effectively_zero( double long x, double long precision )
{
  return abs(x) < precision;
}

float
MathReal::
chop( float x, float tolerance )
{
  return abs(x) < tolerance ? 0.0 : x;
}

double
MathReal::
chop( double x, double tolerance )
{
  return abs(x) < tolerance ? 0.0 : x;
}

double long
MathReal::
chop( double long x, double long tolerance )
{
  return abs(x) < tolerance ? 0.0 : x;
}

void
MathReal::
println( ostream &o, float x )
{
  print(o, x);
  o << '\n';
}

void
MathReal::
println( ostream &o, double x )
{
  print(o, x);
  o << '\n';
}

void
MathReal::
println( ostream &o, double long x )
{
  print(o, x);
  o << '\n';
}

void
MathReal::
printMln( ostream &o, float x )
{
  printM(o, x);
  o << '\n';
}

void
MathReal::
printMln( ostream &o, double x )
{
  printM(o, x);
  o << '\n';
}

void
MathReal::
printMln( ostream &o, double long x )
{
  printM(o, x);
  o << '\n';
}

} // namespace math

#endif // !MathReal_INCLUDED
