// Copyright (C) 2000 Nicholas Schmitt nick@math.umass.edu

#ifndef string_INCLUDED
#define string_INCLUDED

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include <vector.h>
#include "Math.H"

#if defined(__sgi) && !defined(__GNUG__)
#pragma set woff 1375 // destrctor not virtual
#endif


namespace base {

#if defined(__GNUG__)
template<class A>
class StringBase;

template<class A>
ostream &operator<<( ostream &o, StringBase<A> const &s);
#endif

/**
   Class basic_string<>.
   typedef StringBase<char> string;

     Class basic_string<> implements a vector of bytes.

     The methods in in the class StringBase<> below loosely based on the
     SGI basic_string<> class documented at

       http://www.sgi.com/Technology/STL/basic_string.html.

     The implement some but not all of the methods in the SGI class.

     At the bottom of this file is the 'string' typedef, namely

       typedef StringBase<char> string;

   This class is a work in progress and does not yet implement many
   useful string functions.

   Null-terminated strings:
     Class StringBase<> does not implement the method c_str() (returning
     a null-terminated string). To effect the same thing as s.c_str(),
       1. make a temporary copy of your string:
              string tmp(s);
       2. call the method append0() to the copy - this appends a 0 and
          increases the length by 1
              tmp.append0(); (or tmp += 0;)
       3. use the data() method to access a char *.
              e.g. cerr << tmp.data();

   Thread safety:
     I implemented this class as a thread-safe (partial) substitute
     for the SGI StringBase<> class.

     This class subclasses vector<char>, hence is as thread-safe as vector<>.
     In many implementations vector<> is thread-safe if the default allocator
     is used.
*/

template<class A>
class StringBase
  :
  public vector<A>
{

public:
  ///
  static const size_t NPOS;

  /// construct a blank string
  inline StringBase();

  /// construct a string equal to the length-n string s
  inline StringBase( A const s[], size_t n );

  /// construct a string equal to the 0-terminated string s
  inline StringBase( A const s[] );

  /// construct a string equal to the string s
  inline StringBase( StringBase<A> const & );

  // construct a string equal to length-n substring of s
  ///   with offset pos
  StringBase( StringBase<A> const &s, size_t pos, size_t n );

  /// construct a string equal to a single element c
  StringBase( A c );

  /// construct a string filled with n copies of c
  StringBase( size_t n, A c );

  /// destructor
  ~StringBase() {};

  /// set *this to the string s of length n
  void assign( A const s[], size_t n );

  /// set *this to the 0-terminated string s
  void assign( A const s[] );

  /// set *this to the string a
  void assign( StringBase<A> const &a );

  /// set *this to a string of n copies of c
  void assign( size_t n, A c );

  /// set *this to the substing of s starting at pos, of length n
  void assign( StringBase<A> const &s, size_t pos, size_t n );

  /// assign s to *this
  inline StringBase &operator=( A const s[] );

  /// assign s to *this
  inline StringBase &operator=( StringBase<A> const &s );

  /// assign the single element A to *this
  inline StringBase &operator=( A );

  ///
  inline A const *data() const;
  ///
  inline size_t length() const;

  /// append a single A to *this
  inline void append( A c );

  /// append the string s of length n to *this
  void append( A const s[], size_t n );

  /// append the 0-terminated string p to *this
  void append( A const s[] );

  /// append the string s to *this
  void append( StringBase<A> const &s );

  /// append n copies of c to *this
  void append( size_t n, A c );

  /// append A(0) to *this
  inline void append0();

  /// append the 0-terminated string s to *this
  inline void operator+=( A const s[] );

  /// append the string s to *this
  inline void operator+=( StringBase<A> const &s );

  /// append a single A to *this
  inline void operator+=( A );

  /// insert n copies of c befor pos
  void insert( size_t pos, size_t n, A c );

  // replace a substring of *this with s
  ///   namely the length-n substring of *this starting at pos
  void replace( size_t pos, size_t n, StringBase<A> const &s );

  /**
     comparison: returns
       -1 if this < s
       0  if this == s
       1  if this > s
  */
  int compare( A ) const;
  ///
  inline int compare( A const s[] ) const;
  ///
  int compare( A const s[], size_t ) const;
  ///
  int compare( StringBase<A> const & ) const;
  ///
  inline bool operator==( A ) const;
  ///
  inline bool operator==( A const s[] ) const;
  ///
  bool operator==( StringBase<A> const & ) const;
  ///
  inline bool operator!=( A ) const;
  ///
  inline bool operator!=( A const s[] ) const;
  ///
  inline bool operator!=( StringBase<A> const & ) const;
  ///
  inline bool operator<( A ) const;
  ///
  inline bool operator<( A const s[] ) const;
  ///
  inline bool operator<( StringBase<A> const & ) const;
  ///
  inline bool operator>( A ) const;
  ///
  inline bool operator>( A const s[] ) const;
  ///
  inline bool operator>( StringBase<A> const & ) const;
  ///
  inline bool operator<=( A ) const;
  ///
  inline bool operator<=( A const s[] ) const;
  ///
  inline bool operator<=( StringBase<A> const & ) const;
  ///
  inline bool operator>=( A ) const;
  ///
  inline bool operator>=( A const s[] ) const;
  ///
  inline bool operator>=( StringBase<A> const & ) const;

  /// find the first c in *this, starting at position pos
  size_t find( A c, size_t pos = 0 ) const;

  /// find the last c in *this
  size_t rfind( A ) const;

  /// return the length-n substring of *this starting at pos
  inline StringBase<A> substr( size_t pos, size_t n) const;

  StringBase &erase( size_t pos = 0, size_t n = NPOS );

  // write string 's' to stream 'o'
#if !defined(__GNUG__)
  ///
//  friend ostream &operator<<( ostream &o, StringBase<A> const &s);
#endif

#if defined(__GNUG__)
  ///
  friend ostream &operator<< <>( ostream &o, StringBase<A> const &s);
#endif

protected:

private:
  ///
  static size_t stringlength( A const [] );
};

template<class A>
const size_t
StringBase<A>::
NPOS = (size_t)(-1);


template<class A>
void
StringBase<A>::
assign( A const s[], size_t n )
{
  // set *this to the string at p of length n
  clear();

  if ( s == (A *)NULL ) { return; }

  reserve(n);

  size_t i;

  for ( i = 0; i < n; i++ )
  {
    push_back( s[i] );
  }
}


template<class A>
void
StringBase<A>::
assign( A const s[] )
{
  // set *this to the 0-terminated string at p

  clear();

  if( s == (A *)NULL ) { return; }

  size_t i = 0;

  while( s[i] != 0 )
  {
    push_back( s[i] );
    i++;
  }
}


template<class A>
void
StringBase<A>::
assign( StringBase<A> const &a )
{
  // assign a to *this
  if (this != &a )
  {
    // call base-class version
    ((vector<A> *)this)->operator=(a);
  }
}


template<class A>
void
StringBase<A>::
assign( StringBase<A> const &s, size_t pos, size_t n )
{
  // set *this to the substring of s starting at pos, of length n

  if( pos > s.size() ) { return; }

  assign( &s[pos], n );
}


template<class A>
void
StringBase<A>::
assign( size_t n, A c )
{
  // set *this to a string of n copies of c
  reserve( n );

  size_t i;

  for ( i = 0; i < n; i++ )
  {
    push_back( c );
  }
}


template<class A>
StringBase<A>::
StringBase()
    :
    vector<A>()
{
}


template<class A>
StringBase<A>::
StringBase( A const s[] )
    :
    vector<A>()
{
  assign( s );
}


template<class A>
StringBase<A>::
StringBase( A const s[], size_t n )
    :
    vector<A>()
{
  assign( s, n );
}


template<class A>
StringBase<A>::
StringBase( StringBase const &a )
    :
    vector<A>(a)
{
}


template<class A>
StringBase<A>::
StringBase( StringBase const &a, size_t pos, size_t n )
    :
    vector<A>()
{
  // construct a string equal to length-n substring of s
  //   with offset pos
  assign( &a[pos], n );
}


template<class A>
StringBase<A>::
StringBase( size_t n, A c )
    :
    vector<A>()
{
  assign( n, c );
}

template<class A>
StringBase<A>::
StringBase( A c )
    :
    vector<A>()
{
  assign( 1, c );
}


template<class A>
StringBase<A> &
StringBase<A>::
operator=( const A *p )
{
  assign( p );

  return *this;
}


template<class A>
StringBase<A> &
StringBase<A>::
operator=( StringBase<A> const &a )
{
  assign( a );

  return *this;
}


template<class A>
StringBase<A> &
StringBase<A>::
operator=( A c )
{
  clear();

  push_back(c);

  return *this;
}


template<class A>
A const *
StringBase<A>::
data() const
{
  if ( size() == 0 ) { return (A*)NULL; }

  return &operator[](0);
}


template<class A>
size_t
StringBase<A>::
length() const
{
  return size();
}


template<class A>
void
StringBase<A>::
append( A c )
{
  // append a single A to *this
  push_back(c);
}


template<class A>
void
StringBase<A>::
append( A const p[], size_t n )
{
  // append the string starting at p of length n to *this
  size_t i;

  for ( i = 0; i < n; i++ )
  {
    push_back( p[i] );
  }
}


template<class A>
void
StringBase<A>::
append( A const p[] )
{
  // append the 0-terminated string starting at p to *this
  size_t i = 0;

  while( p[i] != 0 )
  {
    push_back( p[i] );
    i++;
  }
}


template<class A>
void
StringBase<A>::
append( StringBase<A> const &a )
{
  // append the string a to *this
  size_t i;

  size_t n = a.size();

  for ( i = 0; i < n; i++ )
  {
    push_back( a[i] );
  }
}


template<class A>
void
StringBase<A>::
append( size_t n, A c  )
{
  // append a string of n c's to *this
  size_t i;

  for ( i = 0; i < n; i++ )
  {
    push_back( c );
  }
}


template<class A>
void
StringBase<A>::
append0( void )
{
  // append a 0 to *this
  append( A(0) );
}


template<class A>
void
StringBase<A>::
operator+=( A c )
{
  // append the single element c to *this
  append( c );
}


template<class A>
void
StringBase<A>::
operator+=( A const s[] )
{
  // append the 0-terminated string p to *this
  append( s );
}


template<class A>
void
StringBase<A>::
operator+=( StringBase<A> const &s )
{
  // append the string s to *this
  append( s );
}


template<class A>
void
StringBase<A>::
replace( size_t pos, size_t n, StringBase<A> const &s )
{
  // replace a substring of *this with s
  // (namely the length-n substring of *this starting at pos)

  if ( size() < pos + n ) { return; }

  vector<A>::iterator i0 = begin() + pos;

  vector<A>::iterator i1 = begin() + pos + n;

  erase( i0, i1 );

  insert( begin() + pos, s.begin(), s.end() );
}


template<class A>
ostream &
operator<<( ostream &o, StringBase<A> const &a)
{
  size_t i;
  for ( i = 0; i < a.size(); i++ )
  {
    o << a[i];
  }
  return o;
}


template<class A>
StringBase<A>
StringBase<A>::substr( size_t pos, size_t n) const
{
  // return the length-n substring of *this starting at pos
  return StringBase<A>( *this, pos, n );
}


template<class A>
int
StringBase<A>::
compare( A c ) const
{
  if ( size() == 0 ) { return -1; };

  A x = operator[](0);

  if ( x < c ) { return -1; }

  if ( x > c ) { return 1; }

  return ( size() == 1 ? 0 : ( size() < 1 ? -1 : 1 ) );
}


template<class A>
int
StringBase<A>::
compare( A const a[] ) const
{
  return compare( a, stringlength(a) );
}


template<class A>
int
StringBase<A>::
compare( A const s[], size_t n ) const
{
  // returns:
  //   -1 if this < s
  //    0 if this == s
  //    1 if this > s

  if ( s == (A *)NULL || n == 0 )
  {
    if ( size() == 0 ) { return 0; }
    else { return 1; }
  }
  else if ( size() == 0 )
  {
    return -1;
  }
  else
  {
    size_t v = (int)Math::min( size(), n);

    size_t i;
    
    for ( i = 0; i < v; i++ )
    {
      if ( operator[](i) < s[i] ) { return -1; }
      if ( operator[](i) > s[i] ) { return 1; }
    }
    
    return ( size() == n ? 0 : ( size() < n ? -1 : 1 ) );
  }
}


template<class A>
int
StringBase<A>::
compare( StringBase<A> const &s) const
{
  // returns -1, 0, 1 according as *this is <, =, > s

  size_t n = Math::min( size(), s.size());

  size_t i;

  for ( i = 0; i < n; i++ )
  {
    if ( operator[](i) < s[i] ) { return -1; }
    if ( operator[](i) > s[i] ) { return 1; }
  }

  return ( size() == s.size() ? 0 : ( size() < s.size() ? -1 : 1 ) );
}


template<class A>
bool
StringBase<A>::
operator==( A const s[] ) const
{
  return compare(s) == 0;
}


template<class A>
bool
StringBase<A>::
operator==( A c ) const
{
  if ( size() != 1 ) { return false; }

  return ( operator[](0) == c );
}


template<class A>
bool
StringBase<A>::
operator==( StringBase<A> const &s) const
{
  if ( size() != s.size() ) { return false; }

  size_t v = size();

  size_t i;

  for ( i = 0; i < v; i++ )
  {
    if ( operator[](i) != s[i] ) { return false; }
  }
  return true;
}


template<class A>
bool
StringBase<A>::
operator!=( A const s[] ) const
{
  return !operator==(s);
}


template<class A>
bool
StringBase<A>::
operator!=( A c ) const
{
  return !operator==(c);
}


template<class A>
bool
StringBase<A>::
operator!=( StringBase<A> const &s ) const
{
  return !operator==(s);
}


template<class A>
bool
StringBase<A>::
operator>( A const s[] ) const
{
  return compare(s) > 0;
}


template<class A>
bool
StringBase<A>::
operator>( A c ) const
{
  return compare(c) > 0;
}


template<class A>
bool
StringBase<A>::
operator>( StringBase<A> const &a) const
{
  return compare(a) > 0;
}


template<class A>
bool
StringBase<A>::
operator<( A const s[] ) const
{
  return compare(s) < 0;
}


template<class A>
bool
StringBase<A>::
operator<( A c ) const
{
  return compare(c) < 0;
}


template<class A>
bool
StringBase<A>::
operator<( StringBase<A> const &a) const
{
  return compare(a) < 0;
}


template<class A>
bool
StringBase<A>::
operator>=( A const s[] ) const
{
  return compare(s) >= 0;
}


template<class A>
bool
StringBase<A>::
operator>=( A c ) const
{
  return compare(c) >= 0;
}


template<class A>
bool
StringBase<A>::
operator>=( StringBase<A> const &a) const
{
  return compare(a) >= 0;
}


template<class A>
bool
StringBase<A>::
operator<=( A const s[] ) const
{
  return compare(s) <= 0;
}


template<class A>
bool
StringBase<A>::
operator<=( A c ) const
{
  return compare(c) <= 0;
}


template<class A>
bool
StringBase<A>::
operator<=( StringBase<A> const &s) const
{
  return compare(s) <= 0;
}


template<class A>
size_t
StringBase<A>::
find( A c, size_t n ) const
{
  if ( n >= size() ) { return NPOS; }

  size_t i;

  for ( i = n; i < size(); i++ )
  {
    if (operator[](i) == c) { return i; }
  }

  return NPOS;
}


template<class A>
size_t
StringBase<A>::
rfind( A c ) const
{
  if ( size() == 0 ) { return NPOS; }

  size_t i = size() - 1;

  for (;;)
  {

    if (operator[](i) == c) { return i; }

    if ( i == 0 ) { break; }

    i--;
  }
  return NPOS;
}


template<class A>
size_t
StringBase<A>::
stringlength( A const s[] )
{
  if ( s==(A *)NULL ) { return 0; }

  size_t i = 0;

  while ( s[i] != 0 ) { i++; }

  return i;
}


template<class A>
StringBase<A> &
StringBase<A>::
erase( size_t pos, size_t n )
{
  if ( pos + n <= size() )
  {
    vector<A>::iterator i = begin();
    vector<A>::erase(i + pos, i + pos + n);
  }
  return *this;
}

template<class A>
void
StringBase<A>::
insert( size_t pos, size_t n, A c )
{
  vector<A>::iterator i = begin();
  vector<A>::insert(i + pos, n, c );
}

} // namespace base

#if defined(__sgi) && !defined(__GNUG__)
#pragma reset woff 1375
#endif

#endif // !string_INCLUDED
