// Copyright (C) 2000 Nicholas Schmitt nick@math.umass.edu

#ifndef vector2_INCLUDED
#define vector2_INCLUDED

#include <iostream.h>
#include <math.h>
#include "Math.H"

namespace math {
using namespace base;

#ifdef __GNUG__
template <class REAL>
class vector2;

template <class REAL>
inline vector2<REAL> operator*( REAL , vector2<REAL> const & );

template <class REAL>
ostream &operator<<( ostream &o, vector2<REAL> const &a );
#endif

/**
  Fixed-length vector class.
  Length N and entries in class REAL.

  file: vector2.H
  */
template <class REAL>
class vector2
{
public:

  /// default constructor
  inline vector2(void);

  ///
  inline vector2( REAL, REAL );

  /// copy constructor
  inline vector2( vector2 const &a );

  /// destructor
  inline virtual ~vector2(void);

  /// assignment operators
  inline vector2<REAL> &operator=( vector2<REAL> const &a );

  /// set to REAL
  inline void set( REAL );

  inline bool operator==( vector2<REAL> const &a );
  inline bool operator!=( vector2<REAL> const &a );
  static inline bool
  equal( vector2<REAL> const &a, vector2<REAL> const &b, REAL epsilon );

  // math methods
  
  /// negation: *this = - *this
  inline void neg(void);

  /// negation: *this = - a
  inline void neg( vector2<REAL> const &a );

  /// *this += a
  inline void add( vector2<REAL> const &a );

  /// *this = a + b
  inline void add( vector2<REAL> const &a, vector2<REAL> const &b );

  /// *this -= a
  inline void sub( vector2<REAL> const &a );

  /// *this = a - b
  inline void sub( vector2<REAL> const &a, vector2<REAL> const &b );

  /// *this *= a
  inline void mul( REAL a );

  /// *this = a * b
  inline void mul( REAL a, vector2<REAL> const &b );

  /// *this = a * b
  inline void mul( vector2<REAL> const &a, REAL b );
  
  /// *this = a + b * c
  inline void mul_add(
    vector2<REAL> const &a, REAL b, vector2<REAL> const &c );

  /// *this += a * b
  inline void mul_add( REAL a, vector2<REAL> const &b );
  
  /// *this /= a
  inline void div( REAL a );

  /// *this = a / b
  inline void div( vector2<REAL> const &a, REAL b );

  /// return dot_product( *this, a )
  static inline REAL
  dot( vector2<REAL> const &a, vector2<REAL> const &b );

  static inline REAL
  det( vector2<REAL> const &a, vector2<REAL> const &b );

  ///
  inline REAL norm(void) const;

  ///
  inline REAL abs(void) const;

  inline void normalize(void);

  // returns perpendicular vector (rotated right 90 degrees)
  inline vector2<REAL> perp(void);

  static inline vector2<REAL> midpoint(
    vector2<REAL> const &,
    vector2<REAL> const & );


  /**
     Returns twice the area of the oriented triangle (a, b, c), i.e., the
     area is positive if the triangle is oriented counterclockwise.
  */
  static inline
  REAL
  det(vector2<REAL> const &a, vector2<REAL> const &b, vector2<REAL> const &c);

  /// Returns true if the points a, b, c are in a counterclockwise order
  static inline bool ccw(
    vector2<REAL> const &a, vector2<REAL> const &b, vector2<REAL> const &c);

  /// Returns true if the points a, b, c are in a clockwise order
  static inline bool cw(
    vector2<REAL> const &a, vector2<REAL> const &b, vector2<REAL> const &c);

  static inline REAL angle( vector2<REAL> const &a, vector2<REAL> const &b );
  static inline REAL angle( vector2<REAL> const &a, vector2<REAL> const &o,
                            vector2<REAL> const &b );


  static bool inCircle(vector2<REAL> const & a, vector2<REAL> const & b,
                       vector2<REAL> const & c, vector2<REAL> const & d);

  // Returns the center of the circle through points a, b, c.
  static vector2<REAL> circumcenter(
    vector2<REAL> const & a,
    vector2<REAL> const & b,
    vector2<REAL> const & c);
  
  static inline
  bool
  same_side(
    vector2<REAL> const &a,
    vector2<REAL> const &b,
    vector2<REAL> const &x,
    vector2<REAL> const &y );

  static inline
  bool
  opposite(
    vector2<REAL> const &a,
    vector2<REAL> const &b,
    vector2<REAL> const &x,
    vector2<REAL> const &y );

  static inline
  bool
  intersect(
    vector2<REAL> const &a1,
    vector2<REAL> const &a2,
    vector2<REAL> const &b1,
    vector2<REAL> const &b2 );

  // returns the intersection of a1<->a2, b1<->b2
  static inline
  vector2<REAL>
  intersection(
    vector2<REAL> const &a1,
    vector2<REAL> const &a2,
    vector2<REAL> const &b1,
    vector2<REAL> const &b2 );


  //------------------------------------------------------------------
  /// operator versions
  
  inline void operator+=( vector2<REAL> const &a ) { this->add(a); }
  ///
  inline void operator-=( vector2<REAL> const &a ) { this->sub(a); }
  ///
  inline void operator*=( REAL a ) { this->mul(a); }
  ///
  inline void operator/=( REAL a ) { this->div(a); }

  //------------------------------------------------------------------
  /// inefficient versions

  inline vector2<REAL> operator+(void) const;
  ///
  inline vector2<REAL> operator-(void) const;
  ///
  inline vector2<REAL> operator+( vector2<REAL> const & ) const;
  ///
  inline vector2<REAL> operator-( vector2<REAL> const & ) const;
  ///
  inline vector2<REAL> operator*( REAL  ) const;

  ///
  inline vector2<REAL> operator/( REAL  ) const;


#ifndef __GNUG__
  friend vector2<REAL> operator*( REAL , vector2<REAL> const & );
#endif

  /// write a to stream o

#ifndef __GNUG__
  friend ostream &operator<<( ostream &o, vector2<REAL> const &a );
#endif

  REAL x, y;
};


// constructors and destructors
template <class REAL>
vector2<REAL>::vector2(void)
    :
    x(0),
    y(0)
{
}

template <class REAL>
vector2<REAL>::vector2( REAL x_, REAL y_ )
    :
    x(x_),
    y(y_)
{
}


template <class REAL>
vector2<REAL>::vector2( vector2<REAL> const &a )
    :
    x(a.x),
    y(a.y)
{
}


template <class REAL>
vector2<REAL>::~vector2(void)
{
}


// assignment operator
template <class REAL>
vector2<REAL> &
vector2<REAL>::operator=( vector2<REAL> const &a )
{
  x = a.x;
  y = a.y;
  return *this;
}


template <class REAL>
void
vector2<REAL>::set( REAL a )
{
  x = a;
  y = a;
}


template <class REAL>
bool
vector2<REAL>::
operator==( vector2<REAL> const &a )
{
  return x == a.x && y == a.y;
}

template <class REAL>
bool
vector2<REAL>::
operator!=( vector2<REAL> const &a )
{
  return x != a.x || y != a.y;
}

template <class REAL>
bool
vector2<REAL>::
equal( vector2<REAL> const &a, vector2<REAL> const &b, REAL epsilon )
{
  REAL x = ::fabs(a.x - b.x);
  if ( x > epsilon ) { return false; }

  REAL y = ::fabs(a.y - b.y);
  if ( y > epsilon ) { return false; }

  return Math::sqr(x) + Math::sqr(y) < Math::sqr(epsilon);
}

// negation: *this = - *this
template <class REAL>
void
vector2<REAL>::neg(void)
{
  x = -x;
  y = -y;
}


// *this = - a
template <class REAL>
void
vector2<REAL>::neg( vector2<REAL> const &a )
{
  x = -a.x;
  y = -a.y;
}


// *this += a
template <class REAL>
void
vector2<REAL>::add( vector2<REAL> const &a )
{
  x += a.x;
  y += a.y;
}


// *this = a + b
template <class REAL>
void
vector2<REAL>::add( vector2<REAL> const &a, vector2<REAL> const &b )
{
  x = a.x + b.x;
  y = a.y + b.y;
}


// *this -= a
template <class REAL>
void
vector2<REAL>::sub( vector2<REAL> const &a )
{
  x -= a.x;
  y -= a.y;
}


// *this = a - b
template <class REAL>
void
vector2<REAL>::sub( vector2<REAL> const &a, vector2<REAL> const &b )
{
  x = a.x - b.x;
  y = a.y - b.y;
}


// *this *= a
template <class REAL>
void
vector2<REAL>::mul( REAL a )
{
  x *= a;
  y *= a;
}


// *this = a * b
template <class REAL>
void
vector2<REAL>::mul( vector2<REAL> const &a, REAL b )
{
  x = a.x * b;
  y = a.y * b;
}


// *this = a * b
template <class REAL>
void
vector2<REAL>::mul( REAL a, vector2<REAL> const &b )
{
  x = a * b.x;
  y = a * b.y;
}

// this /= a
template <class REAL>
void
vector2<REAL>::div( REAL a )
{
  x /= a;
  y /= a;
}


// *this = a / b
template <class REAL>
void
vector2<REAL>::div( vector2<REAL> const &a, REAL b )
{
  x = a.x / b;
  y = a.y / b;
}


// return dot_product( *this, a )
template <class REAL>
REAL
vector2<REAL>::dot( vector2<REAL> const &a, vector2<REAL> const &b )
{
  return a.x * b.x + a.y * b.y;
}

template <class REAL>
REAL
vector2<REAL>::det( vector2<REAL> const &a, vector2<REAL> const &b )
{
  return a.x * b.y - a.y * b.x;
}


// return norm
template <class REAL>
REAL
vector2<REAL>::norm(void) const
{
  return Math::sqr(x) + Math::sqr(y);
}

// return abs
template <class REAL>
REAL
vector2<REAL>::abs(void) const
{
  return sqrt(norm());
}


// return norm
template <class REAL>
void
vector2<REAL>::normalize(void)
{
  REAL r = abs();
  x /= r;
  y /= r;
}

//
template <class REAL>
void
vector2<REAL>::
mul_add( vector2<REAL> const &a, REAL b, vector2<REAL> const &c )
{
  x = a.x + b * c.x;
  y = a.y + b * c.y;
}


//
template <class REAL>
void
vector2<REAL>::
mul_add( REAL a, vector2<REAL> const &b )
{
  x = a * b.x;
  y = a * b.y;
}


// write a to stream o
template <class REAL>
ostream &
operator<<( ostream &o, vector2<REAL> const &a )
{
  o << "{ " << a.x << ", " << a.y << "}";
  return o;
}

template <class REAL>
vector2<REAL>
vector2<REAL>::
midpoint(
  vector2<REAL> const &a,
  vector2<REAL> const &b )
{
  return 0.5*(a+b);
}

//------------------------------------------------------------------
// inefficient versions

template <class REAL>
vector2<REAL>
vector2<REAL>::
operator+(void) const
{
  return *this;
}


//
template <class REAL>
vector2<REAL>
vector2<REAL>::
operator-(void) const
{
  vector2<REAL> tmp(*this);
  tmp.neg();
  return tmp;
}


template <class REAL>
vector2<REAL>
vector2<REAL>::
operator+( vector2<REAL> const &a ) const
{
  vector2<REAL> tmp(*this);
  tmp.add(a);
  return tmp;
}


template <class REAL>
vector2<REAL>
vector2<REAL>::
operator-( vector2<REAL> const &a ) const
{
  vector2<REAL> tmp(*this);
  tmp.sub(a);
  return tmp;
}


template <class REAL>
vector2<REAL>
vector2<REAL>::
operator*( REAL a ) const
{
  vector2<REAL> tmp(*this);
  tmp.mul(a);
  return tmp;
}


template <class REAL>
vector2<REAL>
operator*( REAL a, vector2<REAL> const &b )
{
  vector2<REAL> tmp(b);
  tmp.mul(a);
  return tmp;
}


template <class REAL>
vector2<REAL>
vector2<REAL>::
operator/( REAL a ) const
{
  vector2<REAL> tmp(*this);
  tmp.div(a);
  return tmp;
}


template <class REAL>
REAL
vector2<REAL>::
det(
  vector2<REAL> const &a,
  vector2<REAL> const &b,
  vector2<REAL> const &c)
{
  return (b.x - a.x)*(c.y - a.y) - (b.y - a.y)*(c.x - a.x);
}

template <class REAL>
bool
vector2<REAL>::
ccw(
  vector2<REAL> const &a,
  vector2<REAL> const &b,
  vector2<REAL> const &c)
{
  return det(a, b, c) > 0.0;
}

template <class REAL>
bool
vector2<REAL>::
cw(
  vector2<REAL> const &a,
  vector2<REAL> const &b,
  vector2<REAL> const &c)
{
  return det(a, b, c) < 0.0;
}


template <class REAL>
bool
vector2<REAL>::
inCircle(vector2<REAL> const &a, vector2<REAL> const &b,
         vector2<REAL> const &c, vector2<REAL> const &d)
{
  REAL az = a.norm();
  REAL bz = b.norm();
  REAL cz = c.norm();
  REAL dz = d.norm();

  REAL e = az * det(b, c, d) - bz * det(a, c, d) +
    cz * det(a, b, d) - dz * det(a, b, c);

  return (e > 0.0);
}


template <class REAL>
vector2<REAL>
vector2<REAL>::
circumcenter(
  vector2<REAL> const &a, vector2<REAL> const &b, vector2<REAL> const &c)
{
  // From Graphics Gems I, p.22

  REAL d1, d2, d3, c1, c2, c3;

  d1 = dot( b - a, c - a );
  d2 = dot( b - c, a - c );
  d3 = dot( a - b, c - b );
  c1 = d2 * d3;
  c2 = d3 * d1;
  c3 = d1 * d2;

  return ((c2 + c3)*a + (c3 + c1)*c + (c1 + c2)*b) / (2*(c1 + c2 + c3));
}

template <class REAL>
REAL
vector2<REAL>::
angle( vector2<REAL> const &a, vector2<REAL> const &b )
{
  return atan2( det(a,b), dot(a,b) );
}


template <class REAL>
REAL
vector2<REAL>::
angle( vector2<REAL> const &a, vector2<REAL> const &o,
       vector2<REAL> const &b )
{
  return angle( a-o, b-o );
}

template <class REAL>
vector2<REAL>
vector2<REAL>::
perp()
{
  return vector2<REAL>( -y, x );
}

template <class REAL>
bool
vector2<REAL>::
same_side(
  vector2<REAL> const &a,
  vector2<REAL> const &b,
  vector2<REAL> const &x,
  vector2<REAL> const &y )
{
  // returns true iff x, y are on the same side of the line thru a, b
  return cw(a,b,x) && cw(a,b,y) || cw(b,a,x) && cw(b,a,y);
}

template <class REAL>
bool
vector2<REAL>::
opposite(
  vector2<REAL> const &a,
  vector2<REAL> const &b,
  vector2<REAL> const &x,
  vector2<REAL> const &y )
{
  // returns true iff x, y are on opposite sided of the line thru a, b
  return cw(a,b,x) && cw(b,a,y) || cw(a,b,y) && cw(b,a,x);
}

template <class REAL>
bool
vector2<REAL>::
intersect(
  vector2<REAL> const &a1,
  vector2<REAL> const &a2,
  vector2<REAL> const &b1,
  vector2<REAL> const &b2 )
{
  // returns true iff the segements a1<->a2 and b1<->b2 intersect
  return opposite(a1,a2,b1,b2) && opposite(b1,b2,a1,a2);
}

template <class REAL>
vector2<REAL>
vector2<REAL>::
intersection(
  vector2<REAL> const &a1,
  vector2<REAL> const &a2,
  vector2<REAL> const &b1,
  vector2<REAL> const &b2 )
{
  vector2<REAL> const &a = a2-a1;
  vector2<REAL> const &b = b1-a1;
  vector2<REAL> const &c = b2-a1;

  REAL d = (-b.y*c.x + b.x*c.y) /
    (a.y*b.x - a.x*b.y - a.y*c.x + a.x*c.y);

  vector2<REAL> r;
  r.x = a.x * d;
  r.y = a.y * d;

  return r + a1;
}

} // namespace math

#endif // !vector2_INCLUDED
