// Copyright (C) 2000 GANG http://www.gang.umass.edu

#include <iostream.h>
#include <vector.h>
#include <map.h>
#include "Stl.H"
#include "curve.H"
#include "Edge.H"

namespace curve {

EdgeElement::
EdgeElement()
{
  _prev = (EdgeElement *)NULL;
  _next = (EdgeElement *)NULL;
  _cross = (EdgeElement *)NULL;
  _intersection_point = UNDEFINED_UINT;
  _polygon = UNDEFINED_UINT;
  _orient = NOTHING;
  _face[0] = UNDEFINED_UINT;
  _face[1] = UNDEFINED_UINT;
}


void
EdgeElement::
follow_face(
  EdgeElement const * &e,
  int &direction )
{
  /*
    Suppose f is the face to the
      left of e if direction == 0, or
      right of e if direction == 1.

    Sets e to the next edge of f going counterclockwise along f's boundary
    Sets direction to
      0 iff f is on the left of e
      1 iff f is on the right of e
   */

  if ( direction == 0 )
  {
    e = e->next();
    if (e->orient() == LEFT)
    {
      e = e->cross();
      direction = 0;
    }
    else
    {
      e = e->cross()->prev();
      direction = 1;
    }
  }
  else
  {
    if (e->orient() == LEFT)
    {
      e = e->cross()->prev();
      direction = 1;
    }
    else
    {
      e = e->cross();
      direction = 0;
    }
  }
}


void
EdgeX::
compute() throw ( Exception )
{
  // find the intersection points in the diagram
  compute_intersection_points();

  // compute the orientations
  compute_orientations();

  // find the faces in the diagram
  compute_faces();

  // find the outside face
  compute_outside_face();
}


void
EdgeX::
compute_intersection_points() throw ( Exception )
{
  /*
    Find the intersection points
    Store them three ways:
      (1) vertex_list is a vector, each element of which is a
          pair of uints (a,b). The uints are indices into '_point'
          saying that the segements a<->(a+1) and b<->(b+1) intersect

      (2) vertex_hash is a hash, each element of which is a pair
          of uints (v, e). v=vertex-number, an index into 'vertex_list'.
          Each value of v will occur twice in vertex_hash.
          e=edge-number. When the v's are sorted, these are just 0,1,2,...

    Throws an exception iff the curve drawn by the user with the mouse
    does not parse.
  */

  if ( _polygon.size() < 3 )
  {
    throw ERR_CURVE;
  }

  /*std*/::vector</*std*/::pair<uint, uint> > _vertex_list;
  typedef /*std*/::map< uint, /*std*/::pair<uint,uint> > VertexHash;
  VertexHash _vertex_hash;
#if 0
  /*std*/::vector<uint> _edge;
#endif

  uint c = 0;
  uint i;

  for ( i = 0; i < _polygon.size() - 1; i++ )
  {
    uint j;
    for ( j = i + 1; j < _polygon.size(); j++ )
    {
      uint j1 = ( j == _polygon.size() - 1 ? 0 : j + 1 );
      if ( vector2<real>::
           intersect( _polygon[i], _polygon[i+1],
                      _polygon[j], _polygon[j1] ) )
      {
        _vertex_list.push_back( /*std*/::pair<uint,uint>( i, j ) );
        _vertex_hash[i] = /*std*/::pair<uint,uint>( c, UNDEFINED_UINT );
        _vertex_hash[j] = /*std*/::pair<uint,uint>( c, UNDEFINED_UINT );
        c++;
      }
    }
  }

  if ( _vertex_hash.size() != 2 * _vertex_list.size() )
  {
    throw ERR_CURVE;
  }

  // self-number the hash
  c = 0;
  VertexHash::iterator p;
  for ( p = _vertex_hash.begin(); p != _vertex_hash.end(); p++ )
  {
    (*p).second.second = c++;
  }
  compute_intersection_points( _vertex_list );
  make_edges( _vertex_list, _vertex_hash );
}


void
EdgeX::
compute_intersection_points(
  VertexList const &vertex_list )
{
  uint i;
  for ( i = 0; i != vertex_list.size(); i++ )
  {
    uint a = vertex_list[i].first;
    uint b = vertex_list[i].second;

    uint a2 = next_(a, _polygon);
    uint b2 = next_(b, _polygon);

    Point a0 = _polygon[a];
    Point a1 = _polygon[a2];
    Point b0 = _polygon[b];
    Point b1 = _polygon[b2];

    // compute the R^2 intersection point
    _intersection_point.push_back(
      Point::intersection( a0, a1, b0, b1 ) );
  }
}


void
EdgeX::
make_edges(
  VertexList const &vertex_list,
  VertexHash const &vertex_hash )
{
  // fill _edge with blank edges
  EdgeElement e;
  fill(_edge, e, vertex_hash.size());

  uint i;
  for ( i = 0; i != vertex_list.size(); i++ )
  {
    uint a = vertex_list[i].first;
    uint ax = (*vertex_hash.find(a)).second.second;

    uint b = vertex_list[i].second;
    uint bx = (*vertex_hash.find(b)).second.second;

    _edge[ax]._prev = &_edge[prev_(ax, _edge)];
    _edge[ax]._next = &_edge[next_(ax, _edge)];
    _edge[ax]._cross = &_edge[bx];
    _edge[ax]._intersection_point = i;
    _edge[ax]._polygon = a;

    _edge[bx]._prev = &_edge[prev_(bx, _edge)];
    _edge[bx]._next = &_edge[next_(bx, _edge)];
    _edge[bx]._cross = &_edge[ax];
    _edge[bx]._intersection_point = i;
    _edge[bx]._polygon = b;
  }
}

void
EdgeX::
compute_orientations()
{
  /*std*/::vector<EdgeElement>::iterator i;
  for ( i = _edge.begin(); i != _edge.end(); i++ )
  {
    uint a = (*i)._polygon;
    uint a2 = next_(a, _polygon);

    uint b = (*i).cross()->_polygon;
//    uint b2 = next_(b, _polygon);

    Point a0 = _polygon[a];
    Point a1 = _polygon[a2];
    Point b0 = _polygon[b];
//    Point b1 = _polygon[b2];

    (*i)._orient = ( Point::cw(a0, a1, b0) ?
                     EdgeElement::LEFT :
                     EdgeElement::RIGHT );
  }
}

void
EdgeX::
compute_faces() throw ( Exception )
{
  /*
    COMPUTE FACES

    Compute the _face[] members of _cross.
     _cross[i]_face[j]=n means that the jth face of the cross at
    intersection point i has index n.
    j is 0,1,2,or 3 cyclically
    The indices are non-negative numbers taht consistently number all the
    faces.

    Special case: the diagram is an embedded curve and has no intersection
    points. In this case, _face_count is misleading: there are actually 2
    faces, but _face_count will be set to 0

    ALGORITHM:

    THROWS:
      ERR_FACES if it cannot parse the curve into faces. This
      will happen if the user draws a bad curve with the mouse in at least
      the following ways:
         1. A segement (the line between two mouse sample points) edges
            the curve more than once. This will often happen on the
            final segment which snaps to the beginning of the curve, or
            if the curve is drawn too fast, making long segments
         2. The curve touches or intersects non-transversally (i.e. two
            parts of the curve have a common point and tangent
         3. Three segments intersect at a point
      *** TO DO: this should be made a little more forgiving, e.g. by detecting
      case 1 and bisecting the segment. Also, should detect and report
      more precisely what went wrong.
  */

  _face_count = 0;

  // cycle through the edges
  /*std*/::vector<EdgeElement>::iterator i;
  for ( i = _edge.begin(); i != _edge.end(); i++ )
  {
    uint j;
    for ( j = 0; j != 2; j++ )
    {
      int o = (int)j;
      /*std*/::vector<EdgeElement>::iterator i0 = i;

      if ( (*i0).face(j) == UNDEFINED_UINT )
      {
        do
        {
          if ( (*i0).face(o) != UNDEFINED_UINT ) { throw ERR_FACES; }
          (*i0).face(o) = _face_count;
          EdgeElement const *e = i0;
          EdgeElement::follow_face( e, o );
          i0 = (EdgeElement *)e;
        }
        while ( i0 != i );
        _face_count++;
      }
    }
  }
}


void
EdgeX::
compute_outside_face() throw ( Exception )
{
  /*    
    Returns the index of the "outside" region. The outside region is the
    infinite piece of R^2.

    Algorithm:
      Find the R^2 _point on the curve that is lowest (has the smallest y
      value).
      Starting at this _point, follow the curve (in the direction it was
      drawn) to the next intersection _point.
      At this intersection _point...

    *** TO DO: Probably assumes the curve does not have a cusp at
    the lowest point.
  */

  // CASE: circle (no intersection points): return, with
  // _outside_face set to UNDEFINED_UINT by the constructor
  if (_edge.empty())
  {
    return;
  }

  // sanity check
  // *** TO DO fix this
  if ( _polygon.size() <= 2 )
  {
    throw ERR_INTERNAL;
  }

  uint i;
  uint c = UNDEFINED_UINT;

  // Find the lowest point in R^2 -- i.e. the point (x,y) where
  // y is <= any other point -- by cycling through the all the mouse
  // _points
  real ymin = FLT_MAX;
  for (i = 0; i < _polygon.size(); i++)
  {
    if ( _polygon[i].y < ymin )
    {
      ymin = _polygon[i].y;
      c = i;
    }
  }

  // give up if there is no lowest point
  // (this shouldn't happen)
  if ( c == UNDEFINED_UINT )
  {
    throw ERR_INTERNAL;
  }

  // now c is the index (into _points) of the minimum

  // Compute the orientation of the curve relative to the outside region:
  //   orientation == 1: curve moves to the right at its minimum
  //   orientation == 0: curve moves to the left at its minimum
  uint c1 = next_(c, _polygon);
  real x = _polygon[c1].x - _polygon[c].x;

  // give up if the curve does not move horizontally at its minimum
  if ( x == 0 )
  {
    throw ERR_INTERNAL;
  }

  int orientation = ( x > 0 ? 1 : 0 );

  // find the edges before and after the lowest point

  uint e = UNDEFINED_UINT;
  for ( i = 0; i < _edge.size(); i++ )
  {
    uint p = _edge[i].polygon();
    uint q = _edge[next_(i, _edge)].polygon();
    if ( p < c && c < q || q < p && (p < c || c < q) )
    {
      e = i;
      break;
    }
  }

  // *** TO DO: This search does not allow for the possibility that the
  // lowest point is an intersection point
  if ( e == UNDEFINED_UINT )
  {
    throw ERR_CURVE;
  }

  // _edge is now the index (into _edge) of the _edge immediately
  // after the minimum point

  _outside_face = _edge[i].face(orientation);
}

void
EdgeX::
dump( /*std*/::ostream &o )
{
  /*std*/::cerr << "EDGE DUMP\n";
  uint c = 0;
  /*std*/::vector<EdgeElement>::iterator i;
  for ( i = _edge.begin(); i != _edge.end(); i++ )
  {
    o << "edge " << c++ << " face "
      << (*i).face(0) << " face "
      << (*i).face(1) << /*std*/::endl;
  }
}

} // namespace curve
