// Copyright (C) 2000 GANG http://www.gang.umass.edu

#include <iostream.h>
#include <iomanip.h>
#include "Stl.H"
#include "curve.H"
#include "Face.H"

namespace curve {

void
FaceX::
compute() throw ( Exception )
{
  compute_faces();
  compute_degrees();
  compute_max_degree();
  construct();
  compute_euler_characteristic();
}

void
FaceX::
compute_faces()
{
  /*
    Compute faces from edges
  */
  FaceElement face;
  face.degree = UNDEFINED_INT;
  fill( _face, face, _edge.face_count() );

  // cycle through the edges
  /*std*/::vector<EdgeElement>::const_iterator i;
  for ( i = _edge.edge().begin(); i != _edge.edge().end(); i++ )
  {
    uint j;
    for ( j = 0; j != 2; j++ )
    {
      FaceElement &f = _face[(*i).face(j)];
      if ( f.edge.empty() )
      {
        int o = (int)j;
        /*std*/::vector<EdgeElement>::const_iterator i0 = i;
        
        do
        {
          FaceElement::EdgeOfFace q;
          q.edge = i0 - &_edge[0];
          q.opp_face = UNDEFINED_UINT;
          q.opp_edge = UNDEFINED_UINT;
          f.edge.push_back( q );
          EdgeElement::follow_face( i0, o );
        }
        while ( i0 != i );
      }
    }
  }
}

void
FaceX::
compute_degrees() throw ( Exception )
{
  /*
    Number the faces of the diagram by "degree." In the case that the
    curve bounds an immersed surface, these are the degrees of the
    immersion at each face, i.e. the number of preimages.
  */

  int n = 0;

  if (_face.size() == 0 )
  {
    _has_degrees = TRUE;
    return;
  }

  // cycle through edges
  uint i;
  for (i = 0; i != _edge.size(); i++)
  {
    n += ( _edge[i].orient() == EdgeElement::LEFT ? -1 : 1 );
    int &a = _face[_edge[i].face(0)].degree;
    int &b = _face[_edge[i].face(1)].degree;
    if ( a != UNDEFINED_INT && a != n || b != UNDEFINED_INT && b != n - 1 )
    {
      throw ERR_DEGREE;
    }
    a = n;
    b = n - 1;
  }

  int d = _face[_edge.outside_face()].degree;

  // Shift the degrees so that the outside region has degree 0.
  for ( i = 0; i != _face.size(); i++ )
  {
    _face[i].degree -= d;
  }

  compute_has_degrees();
}

void
FaceX::
compute_has_degrees()
{
  // set _has_degrees to true iff all degrees are non-negative
  _has_degrees = TRUE;

  uint i;
  for (i = 0; i != _face.size(); i++)
  {
    if (_face[i].degree < 0 )
    { _has_degrees = FALSE; break; }
  }
}


bool
FaceX::
has_degrees() const
{
  return (bool)_has_degrees;
}

void
FaceX::
compute_max_degree()
{
  /*
    Set '_max_degree' to maximum of the face degrees.
  */

  // case of no faces (trivial curve with no intersection points):
  // return with max_degree == UNDEFINED_UINT
  if (_face.empty() )
  {
    return;
  }

  _max_degree = -INT_MAX;
  /*std*/::vector<FaceElement>::const_iterator q;
  for ( q = _face.begin(); q != _face.end(); q++ )
  {
    if ( (*q).degree > _max_degree ) { _max_degree = (*q).degree; }
  }
  if ( _max_degree == -INT_MAX )
  {
    throw ERR_INTERNAL;
  }
}


void
FaceX::
construct()
{
  // the curve must have non-negative degrees for this function to work
  // properly

  //  if (!_has_degrees) { return; }

  // initialize the opp_face element
  /*std*/::vector<FaceElement>::iterator i;
  for ( i = _face.begin(); i != _face.end(); i++ )
  {
    /*std*/::vector<FaceElement::EdgeOfFace>::iterator j;
    for ( j = (*i).edge.begin(); j != (*i).edge.end(); j++ )
    {
      uint e = (*j).edge;
      if ( _edge[e].face(0) == _edge[e].face(1) )
      {
        throw ERR_INTERNAL;
      }

      uint c = ( _edge[e].face(0) == (uint)(i - &_face[0UL]) ? 1UL : 0UL );

      uint f = _edge[e].face(c);

      (*j).opp_face = f;
    }
  }

  // initialize the opp_edge element
  for ( i = _face.begin(); i != _face.end(); i++ )
  {
    /*std*/::vector<FaceElement::EdgeOfFace>::iterator j;
    for ( j = (*i).edge.begin(); j != (*i).edge.end(); j++ )
    {
      // find corresponding face-edge
      FaceElement const *p = &_face[(*j).opp_face];
      /*std*/::vector<FaceElement::EdgeOfFace>::const_iterator k;
      for ( k = p->edge.begin(); k != p->edge.end(); k++ )
      {
        if ( (*k).edge == (*j).edge ) { break; }
      }
      if ( k == p->edge.end() )
      {
        throw ERR_INTERNAL;
      }
      uint c = k - p->edge.begin();
      (*j).opp_edge = c;
    }
  }
}


void
FaceX::
compute_euler_characteristic()
{
  /*
    Computes the euler characteristic

           Euler characteristic = vertices - edges + faces

    of the domain of the immersion by counting  vertices, edges and faces
    with the correct multiplicity.

    Vertices: Counts with multiplicity n a vertex with degree numbers
      n, n-1, n-1, n-2.

             |
        n-1  |  n
        -----+-----     counts as n vertices
        n-2  | n-1
             |

    Edges: Counts with multiplicity n an edge with degree numbers n, n-1.

             n
         ---------     counts as n edges
            n-1


    FaceElements: Counts as n a face with degree number n

         +-------+
         |       |
         |   n   |    counts as n faces
         |       |
         +-------+

    This calculation assumes that the boundary curve is connected.

  */

  // CASE: trivial curve (no intersection points): return with
  // euler_characteristic == UNDEFINED_INT
  if (_face.empty())
  {
    _euler_characteristic = 1;
    return;
  }

  uint i;

  // Count the number of vertices of the preimage of the immersion
  //   with correct multiplicity by picking the corner with greatest
  //   multiplicity.
  // Counts each vertex twice (once for each edge going through it)
  //   and divides by 2 at the end.
  int vertex = 0;
  for ( i = 0; i < _edge.size(); i++ )
  {
    int d = _face[_edge[i].face(0)].degree;
    if ( _edge[i].orient() == 0 )
    {
      vertex += d;
    }
    else
    {
      vertex += d + 1;
    }
  }
  if ( vertex % 2 != 0 )
  {
    throw ERR_INTERNAL;
  }
  vertex /= 2;


  // Edges: cycle through the edges. For each edge, add the larger degree of
  //    the edges two adjacent faces
  int edge = 0;
  for ( i = 0; i < _edge.size(); i++ )
  {
    edge += Math::max( _face[_edge[i].face(0)].degree,
                       _face[_edge[i].face(1)].degree );
  }


  // FaceElements: add up the degrees of all the faces.
  int face = 0;
  for (i = 0; i < _face.size(); i++ )
  {
    face += _face[i].degree;
  }

  // Return the Euler characteristic
  _euler_characteristic = vertex - edge + face;
}


void
FaceX::
dump( /*std*/::ostream &o ) const
{
  uint c = 0;
  /*std*/::cerr << "FACE DUMP\n";
  /*std*/::vector<FaceElement>::const_iterator i;
  for ( i = _face.begin(); i != _face.end(); i++ )
  {
    o << "face " << c++ << ": degree=" << (*i).degree << " ";
    uint j;
    for ( j = 0; j != (*i).edge.size(); j++ )
    {
      o << "[" << (int)(*i).edge[j].edge << " "
           << (int)(*i).edge[j].opp_face << " "
           << (int)(*i).edge[j].opp_edge << "] ";
    }
    o << /*std*/::endl;
  }
}

void
FaceX::
dump_degrees( /*std*/::ostream &o ) const
{
  // report degrees
  if ( !_face.empty() )
  { 
    o << "DEGREES:\n";
    /*std*/::vector<FaceElement>::const_iterator i;
    for ( i = _face.begin(); i != _face.end(); i++ )
    {
      o << "face " << /*std*/::setw(2) << i - _face.begin()
        << " has degree " << (*i).degree << "\n";
    }
    o.flush();
  }
}

void
FaceX::
add_degree( int d )
{
  /*std*/::vector<FaceElement>::iterator i;
  for ( i = _face.begin(); i != _face.end(); i++ )
  {
    (*i).degree += d;
  }
  compute_has_degrees();
  compute_max_degree();
  compute_euler_characteristic();
}

} // namespace curve
