// Copyright (C) 2000 GANG http://www.gang.umass.edu

#include <iostream.h>
#include <iomanip.h>
#include <algo.h>
#include <list.h>
#include <vector.h>
#include "Use.h"
#include "Stl.H"
#include "curve.H"
#include "Graph.H"

namespace curve {

//#define SEARCH_TRACE

bool
Graph::
next_connection(
  FaceDotEdge &y ) const
{
  /*
    Increments opp_dot to next dot that connects.
    Returns: true: if a connection is found
             false: there are no more connections
   */

  // check for the case of all dots on this edge are UNDEFINED_UINT
  // this prevents redundently choosing dots which are unconnected
  if (y.dot != UNDEFINED_UINT )
  {
    bool flag = true;
    uint i;
    for ( i = 0; i != _face[y.face].edge.size(); i++ )
    {
      if ( _graph[y.face][y.dot][i] != UNDEFINED_UINT )
      {
        flag = false;
        break;
      }
    }
    if (flag == true) { return false; }
  }

  // increment dot
  y.dot = ( y.dot == UNDEFINED_UINT ? 0 : y.dot + 1 );

  // sanity check
  if ( y.face >= _graph.size() )
  {
    throw ERR_INTERNAL;
  }

  // look through the dots in the adacent face for a free dot
  uint n = _graph[y.face].size();

  while(
    y.dot != n &&
    _graph[y.face][y.dot][y.edge] != UNDEFINED_UINT )
  {
    y.dot++;
  }

  // return: true if y.dot < n
  //         false if y.dot == n
  return y.dot != n;
}

void
Graph::
choose_connection(
  FaceDotEdge &x, FaceDotEdge &y,
  /*std*/::vector</*std*/::pair<uint,uint> > const &face_list ) const
{
  /*
    find a face=dot-edge to be connected

    ** TO DO: this is inefficient
  */

  x.face = UNDEFINED_UINT;
  x.edge = UNDEFINED_UINT;
  x.dot = UNDEFINED_UINT;
  y.face = UNDEFINED_UINT;
  y.edge = UNDEFINED_UINT;
  y.dot = UNDEFINED_UINT;

  if (_face.max_degree() == UNDEFINED_INT) { return; }

  // cycle through _sorted_face_list
  /*std*/::vector</*std*/::pair<uint, uint> >::const_iterator i;
  for ( i = face_list.begin();
        i != face_list.end(); i++ )
  {
    uint e = (*i).second;

    // compute c = number of graph connections across this edge
    uint c = 0;

    FaceDot const &g = _graph[(*i).first];

    // compute k0 = last dot connected across edge
    FaceDot::const_iterator k0 = g.begin();

    // cycle through the dots of the face
    FaceDot::const_iterator k;
    for (k = g.begin(); k != g.end(); k++ )
    {
      if ( (*k)[e] != UNDEFINED_UINT )
      { k0 = k+1; c++; }
    }

    FaceElement const &f = _face[(*i).first];
    if ( (int)c < f.degree )
    {
      // found an unconnected face-dot-edge
      x.face = (*i).first;
      x.edge = e;
      x.dot = k0 - g.begin();
      y.face = f.edge[e].opp_face;
      y.edge = f.edge[e].opp_edge;
      break;
    }
  }
}


void
Graph::
insert_edge(
  FaceDotEdge const &x, FaceDotEdge const &y )
{
  /*
    Adds an edge to a connection graph
  */

  DIE_IF(_graph[x.face][x.dot][x.edge] != UNDEFINED_UINT);
  DIE_IF(_graph[y.face][y.dot][y.edge] != UNDEFINED_UINT);

  // add the edge to the graph
  _graph[x.face][x.dot][x.edge] = y.dot;
  _graph[y.face][y.dot][y.edge] = x.dot;

  // counts

  // x count
  DIE_IF( _connection_count[x.face][x.edge] == 0 );
  _connection_count[x.face][x.edge]--;

  // y count
  DIE_IF( _connection_count[y.face][y.edge] == 0 );
  _connection_count[y.face][y.edge]--;
}

void
Graph::
remove_edge(
  FaceDotEdge const &x, FaceDotEdge const &y )
{
  /*
    Adds an edge to a connection graph
  */

  DIE_IF(_graph[x.face][x.dot][x.edge] == UNDEFINED_UINT);
  DIE_IF(_graph[y.face][y.dot][y.edge] == UNDEFINED_UINT);

  // add the edge to the graph
  _graph[x.face][x.dot][x.edge] = UNDEFINED_UINT;
  _graph[y.face][y.dot][y.edge] = UNDEFINED_UINT;

  // counts

  // x count
  _connection_count[x.face][x.edge]++;

  // y count
  _connection_count[y.face][y.edge]++;
}

bool
Graph::
diamond_check( FaceDotEdge const &x, FaceDotEdge const &y ) const
{
  /*
    Returns true iff a partially completed graph does not break the
    "diamond rule" with face/dot/edge as one of the diamond edges
  */

  bool r = diamond_check_edge( x, y );
  if ( !r ) { return false; }

  // think of the new graph-edge in the opposite orientation
  r = diamond_check_edge( y, x );

  return r;
}


bool
Graph::
diamond_check_edge(
  FaceDotEdge const &x, FaceDotEdge const &y ) const
{
  /*
    Returns true iff a new graph edge, if added to a partially completed
    graph, does not break the "diamond rule".

    Given a face 'face', a dot 'dot' in 'face', and an edge 'edge'
    of 'face', checks whether adding a graph-edge from 'dot' in 'face'
    across 'edge' breaks the diamond rule.

    a3<-||---a2           b0-- ||-> b1
        ||   ^                 ||   |
    ============          ============
        ||   |                 ||   v
    a0- || ->a1           b3<- ||---b2


   *** TO DO: check for forced edges.
  */


  // Compute face/dot/edge for four vertices ccw from the given face/dot/edge
  // Put this info into a[0],a[1],a[2],a[3]

  uint i;

  FaceDotEdge a[4];

  a[0] = y;
  a[0].edge = prev_( a[0].edge, _face[a[0].face].edge );

  for (i = 1; i != 4; i++ )
  {
    around( a[i], a[i-1], true );
    if ( a[i].dot == UNDEFINED_UINT ) { break; }
  }
  uint an = i - 1;

  // Compute face/dot/edge for four vertices ccw from the given face/dot/edge
  // Put this info into b[0],b[1],b[2],b[3]

  FaceDotEdge b[4];

  b[0] = x;
  b[0].edge = next_( b[0].edge, _face[b[0].face].edge );

  for (i = 1; i != 4; i++ )
  {
    around( b[i], b[i-1], false );
    if ( b[i].dot == UNDEFINED_UINT ) { break; }
  }
  uint bn = i - 1;

  // Perform the checks

  bool r = true;

  uint c = an + bn;

  // c+1 = the number of complete edges of the diamond made by the new edge
  if (c == 0)
  {
    // 1-sided diamond
    // the vertices of the new edge are not connnected to anything:
    // return true
  }
  else if (c == 1)
  {
    // 2-sided diamond
    // the new edge makes an "L" with one other edge
    // Check if the diamond can be completed:
    //   If the diamond cannot be completed return false;
    //   If the diamond can be completed uniquely, do so and return true
    //   If the diamond can be completed non-uniquely, return true
    if ( an == 0 )
    {
      uint c0 = _connection_count[b[1].face][b[1].edge];
      uint c1 = _connection_count[a[0].face][a[0].edge];
      if ( c0 == 0 || c1 == 0 )
      {
        // no connection can be made
        r = false;
      }
#if 0
      // This never happens because of the increasing degree-order
      else if ( c0 == 1 && c1 == 1 )
      {
        // make connection
        /*std*/::cerr << "CONNECT\n";
      }
#endif
    }
    else if ( an == 1 )
    {
      uint c0 = _connection_count[a[1].face][a[1].edge];
      uint c1 = _connection_count[b[0].face][b[0].edge];
      if ( c0 == 0 || c1 == 0 )
      {
        r = false;
      }
#if 0
      // This never happens because of the increasing degree-order
      else if ( c0 == 1 && c1 == 1 )
      {
        /*std*/::cerr << "CONNECT\n";
      }
#endif
    }
  }
  else if (c == 2)
  {
    // 3-sided diamond
    // Check if the diamond can be completed
    if ( an == 0 )
    {
      uint c0 = _connection_count[a[0].face][a[0].edge];
      if ( c0 == 0 )
      {
        r = false;
      }
      else
      {
//        /*std*/::cerr << "CONNECT\n";
      }
    }
    else if ( an == 1 )
    {
      uint c0 = _connection_count[a[1].face][a[1].edge];
      if ( c0 == 0 )
      {
        r = false;
      }
      else
      {
//        /*std*/::cerr << "CONNECT\n";
      }
    }
    else if ( an == 2 )
    {
      uint c0 = _connection_count[a[2].face][a[2].edge];
      if ( c0 == 0 )
      {
        r = false;
      }
      else
      {
//        /*std*/::cerr << "CONNECT\n";
      }
    }
  }
  else if (c == 3 || c == 4 || c == 5)
  {
    // contradictory diamond with 4 or more sides
    r = false;
  }
  else if (c == 6)
  {
    // diamond is complete
    if ( a[3].dot != b[0].dot )
    {
      // contradictory diamond
      r = false;
    }
    if (r)
    {
      DIE_IF( a[0].dot != b[3].dot );
      DIE_IF( a[1].dot != b[2].dot );
      DIE_IF( a[2].dot != b[1].dot );
      DIE_IF( a[3].dot != b[0].dot );
    }
  }
  else
  {
    DIE();
  }

#if 0
  // DUMP
  {
  uint i;
  for (i = 0; i != 4; i++ )
  {
    cerr << "check a[" << i << "]=" << a[i] << "\n";
  }
  for (i = 0; i != 4; i++ )
  {
    cerr << "check b[" << i << "]=" << b[i] << "\n";
  }
  cerr << "an=" << an << " bn=" << bn << " r=" << r << "\n\n";
  
  }
#endif

  return r;
}


bool
Graph::
diamond_check() const
{
  /*
    Returns true iff a completed graph does not break the "diamond rule".
  */

  bool r = true;

  // cycle through faces of graph
  /*std*/::vector<FaceDot>::const_iterator i;
  for ( i = _graph.begin(); i != _graph.end(); i++ )
  {
    // cycle through dots of face
    /*std*/::vector<Dot>::const_iterator j;
    for ( j = (*i).begin(); j != (*i).end(); j++ )
    {
      // cycle through edges of dot
      /*std*/::vector<uint>::const_iterator k;
      for ( k = (*j).begin(); k != (*j).end(); k++ )
      {
        if ( *k != UNDEFINED_UINT )
        {
          // set k1 to the edge after k
          /*std*/::vector<uint>::const_iterator k1 = k + 1;
          if ( k1 == (*j).end() ) { k1 = (*j).begin(); }

          // if a dot is connected across two consecutive edges
          if ( *k1 != UNDEFINED_UINT )
          {
            // check for diamond
            r = diamond_check_edge1(
              FaceDotEdge( i - _graph.begin(),
                           j - (*i).begin(),
                           k - (*j).begin() ) );

            if ( !r ) { goto end; }
          }
        }
      }
    }
  }
  end:
  return r;
}

void
Graph::
around( FaceDotEdge &b, FaceDotEdge const &a, bool ccw ) const
{
  /*
    Sets b to the next or prev face/dot/edge from a
  */

  // set b.dot to the dot across face/dot/edge of a[i-1]
  b.dot = _graph[a.face][a.dot][a.edge];
  
  if ( b.dot == UNDEFINED_UINT ) { return; }

  // set b.face to the face across face/dot/edge of a[i-1]
  b.face = _face[a.face].edge[a.edge].opp_face;
  
  // set edge to the edge cw from opp_edge of face/dot/edge of a[i-1]
  uint edge = _face[a.face].edge[a.edge].opp_edge;
  b.edge = ccw ? prev_(edge, _face[b.face].edge) :
    next_(edge, _face[b.face].edge);
}


int
Graph::
diamond_check_edge1( FaceDotEdge const &x0 ) const
{
  /*
    Checks that given two diamond edges, the whole diamond exists.


    x0  ||                y0-- ||-> y1
     |  ||                     ||   |
    ============          ============
     v  ||                     ||   v
    x1- || ->x2                ||  y2

  */ 
  bool r = true;

  FaceDotEdge x1, x2;
  around( x1, x0, true );
  DIE_IF( x1.dot == UNDEFINED_UINT );
  around( x2, x1, true );

  FaceDotEdge y0, y1, y2;

  // Perform the check
  if ( x2.dot == UNDEFINED_UINT )
  {
    r = false;
    goto end;
  }

  y0 = x0;
  y0.edge = next_(y0.edge, _face[y0.face].edge);

  around( y1, y0, false );
  DIE_IF( y1.dot == UNDEFINED_UINT );
  around( y2, y1, false );

  // Perform the check
  if ( y2.dot == UNDEFINED_UINT )
  {
    r = false;
    goto end;
  }

  // x2.dot should equal y2.dot because this is the check of a complete
  // graph, every edge of which passes the diamond test
  DIE_IF( x2.dot != y2.dot );

  end:
  return r;
}


void
Graph::
dump( /*std*/::ostream &o ) const
{
  if ( _face.empty() ) { return; }

  // cycle through faces
  /*std*/::vector<FaceElement>::const_iterator i;
  for ( i = _face.face().begin(); i != _face.face().end(); i++ )
  {
    int d = (*i).degree;
    uint i0 = i - &_face[0];
    // cycle through edges
    uint j;
    for ( j = 0; j != (*i).edge.size(); j++ )
    {
      // report only those edges for which current face *i has higher degree
      if ( d > _face[(*i).edge[j].opp_face].degree )
      {
        // cycle through dots
        uint k;
        for ( k = 0; k != _graph[i0].size(); k++ )
        {
          if ( _graph[i0][k][j] != UNDEFINED_UINT )
          {
            o << "face " << i0
                 << " dot " << k
                 << " across edge "
                 << /*std*/::setw(2) << (*i).edge[j].edge
                 << " to face " << (*i).edge[j].opp_face
                 << " dot " << _graph[i0][k][j]
                 << /*std*/::endl;
          }
        }
      }
    }
  }
}


bool
Graph::
same(
  Graph const &graph1 ) const
{
  /*
    Returns true iff graph1 and graph2 are the same up to permutations
    of dots in each face.

    A DictElement is a translation table for the dots in a single face,
    from one graph to another.
        done: flag saying we have a translation table for this face
        dot: the translation table
    We need a vector<DictElement> to store the traslation of two whole
      graphs.

    ALGORITHM:
      Construct a "dictionary" relating the dots in graph0 to those in
      graph1. Do the faces in degree order. Each face of lower degree
      provides the dictionary for the face of the next order. An exception
      is "orphan" faces -- faces of degree n > 0 with no adjacent face
      of degree n-1. These are handled by using the adjacent face of
      higher degree, presumed to have dictionary entries already.
  */

  // initialize permutation dictionary
  /*std*/::vector<DictElement> dict;
  uint i;
  for ( i = 0; i != _face.size(); i++ )
  {
    /*std*/::vector<uint> dot;
    fill( dot, UNDEFINED_UINT, _face[i].degree );
    DictElement d;
    d.done = (_face[i].degree == 0 ? TRUE : FALSE);
    d.dot = dot;
    dict.push_back( d );
  }

  // return value
  bool r = TRUE;

  uint d;
  for (d = 0; d < (uint)_face.max_degree(); d++ )
  {
    /*std*/::vector<FaceElement>::const_iterator i;
    for ( i = _face.face().begin(); i != _face.face().end(); i++ )
    {
      if ( (uint)(*i).degree == d )
      {
        uint face = i - &_face[0];

        // cycle through edges of face i
        // find adjacent faces with degree c+1
        uint j;
        for ( j = 0; j != (*i).edge.size(); j++ )
        {
          uint opp_face = (*i).edge[j].opp_face;
          if ( (uint)_face[opp_face].degree == d + 1 )
          {
            r = dictionary_entry( graph1, face, j, dict );
            if (!r) { goto end; }
          }
        }
      }
    }
  }

  // try to complete the dictionary for the not-done faces
  for ( i = 0; i != _face.size(); i++ )
  {
    if (! dict[i].done )
    {
      // cycle through the edges of the not-done face
      uint j;
      for ( j = 0; j != _face[i].edge.size(); j++ )
      {
        r = dictionary_entry2( graph1, i, j, dict );
        if (!r) { goto end; }
      }
    }
  }

  // Check whether dictionary was completed successfully
  for ( i = 0; i != _face.size(); i++ )
  {
    if (! dict[i].done )
    {
      /*std*/::cerr << "*** could not compare graphs\n";
      r = false;
      goto end;
    }
  }
#if 1
  // Check the consistency of the dictionary
  // (Is this necessary?)
  {
    // cycle through faces
    uint i;
    for ( i = 0; i != _face.size(); i++ )
    {
      // cycle through edges
      uint edge;
      for ( edge = 0; edge != _face[i].edge.size(); edge++ )
      {
        uint opp_face = _face[i].edge[edge].opp_face;

        // cycle through dots
        uint dot;
        for ( dot = 0; dot != (uint)_face[i].degree; dot++ )
        {
          // look up the dot in the dictionary
          uint dot1 = dict[i].dot[dot];
          uint d = _graph[i][dot][edge];
          uint d1 = graph1._graph[i][dot1][edge];

          // d1 should be the dictionary translate of d

          if ( d == UNDEFINED_UINT && d1 == UNDEFINED_UINT ||
               d1 == dict[opp_face].dot[d] )
          {
          }
          else
          {
#if 0
            /*std*/::cerr << "dictionary is wrong\n";
#endif
            r = false;
            goto end;
          }
        }
      }
    }
  }
#endif
  end:

#if 0
  if (r)
  {
    // Dump the dictionary
    /*std*/::cerr << "----------\nDICTIONARY:\n";
    for ( i = 0; i != _face->size(); i++ )
    {
      /*std*/::cerr << "face " << i << " " << " status=" << dict[i].done << " ";
      uint j;
      for ( j = 0; j != dict[i].dot.size(); j++ )
      {
        /*std*/::cerr << "(" << j << "," << (int)dict[i].dot[j] << ")";
      }
      /*std*/::cerr << /*std*/::endl;
    }
  }
#endif

  return r;
}


bool3
Graph::
dictionary_entry(
  Graph const &graph1,
  uint face,
  uint edge,
  /*std*/::vector<DictElement> &dict ) const
{
  /*
    Used by graphs_are_same()
  */

  bool3 r = TRUE;
  uint k;
  uint k0, k1;
  uint d;

  uint face1 = _face[face].edge[edge].opp_face;

  if ( face1 >= dict.size() )
  {
    throw ERR_INTERNAL;
  }

  uint edge1 = _face[face].edge[edge].opp_edge;

  if ( _face[face].degree > _face[face1].degree )
  {
    throw ERR_INTERNAL;
  }

  if ( dict[face1].done )
  {
    goto end;
  }


  d = (uint)_face[face].degree;

  // Assign a dictionary entry to the unconnected dot in face1

  // look for the undefined dot in each graph
  k0 = UNDEFINED_UINT;
  for ( k = 0; k != d + 1; k++ )
  {
    if ( _graph[face1][k][edge1] == UNDEFINED_UINT )
    {
      k0 = k;
      break;
    }
  }
  if ( k0 == UNDEFINED_UINT )
  {
    throw ERR_INTERNAL;
  }

  k1 = UNDEFINED_UINT;
  for ( k = 0; k != d + 1; k++ )
  {
    if ( graph1._graph[face1][k][edge1] == UNDEFINED_UINT )
    {
      k1 = k;
      break;
    }
  }
  if ( k1 == UNDEFINED_UINT )
  {
    throw ERR_INTERNAL;
  }


  if ( dict[face1].dot[k0] == UNDEFINED_UINT )
  {
    dict[face1].dot[k0] = k1;
  }
  else
  {
    if ( dict[face1].dot[k0] != k1 )
    {
      r = FALSE;
      goto end;
    }
  }

  // assign dictionary entries for other dots in face1

  if ( dict[face].done )
  {
    // cycle through the dots in the degree c face 'face'
    // assign dictionary entries
    uint k;
    for ( k = 0; k != d; k++ )
    {
      // insert new entry in dict

      // lookup d0 in the dictionary
      // set d to the corresponding dot in graph1
      uint d = dict[face].dot[k];
      if ( d == UNDEFINED_UINT )
      {
        throw ERR_INTERNAL;
      }

      // d0 is the opposite dot in graph0
      uint d0 = _graph[face][k][edge];

      // d1 is the opposite dot in graph1
      uint d1 = graph1._graph[face][d][edge];

      if ( dict[face1].dot[d0] == UNDEFINED_UINT )
      {
        dict[face1].dot[d0] = d1;
      }
      else
      {
        if (dict[face1].dot[d0] != d1)
        {
          r = FALSE;
          goto end;
        }
      }
    }
    dict[face1].done = TRUE;
  }
  end:
  return r;
}


bool3
Graph::
dictionary_entry2(
  Graph const &graph1,
  uint face,
  uint edge,
  /*std*/::vector<DictElement> &dict ) const
{
  /*
    Used by graphs_are_same()
  */

  bool3 r = TRUE;
  uint d;

  uint face1 = _face[face].edge[edge].opp_face;

  if ( face1 >= dict.size() )
  {
    throw ERR_INTERNAL;
  }

  uint edge1 = _face[face].edge[edge].opp_edge;

  if ( _face[face].degree > _face[face1].degree )
  {
    throw ERR_INTERNAL;
  }

  if ( dict[face].done )
  {
    goto end;
  }
  if ( ! dict[face1].done )
  {
    goto end;
  }

  d = (uint)_face[face].degree;

  // assign dictionary entries for other dots in face1

  // cycle through the dots in the degree c face 'face'
  // assign dictionary entries
  uint k;
  for ( k = 0; k != d + 1; k++ )
  {
    // insert new entry in dict

    // lookup d0 in the dictionary
    // set d to the corresponding dot in graph1
    uint d = dict[face1].dot[k];
    if ( d == UNDEFINED_UINT )
    {
      throw ERR_INTERNAL;
    }

    // d0 is the opposite dot in graph0
    uint d0 = _graph[face1][k][edge1];

    // d1 is the opposite dot in graph1
    uint d1 = graph1._graph[face1][d][edge1];

    if ( d0 == UNDEFINED_UINT && d1 != UNDEFINED_UINT ||
         d0 != UNDEFINED_UINT && d1 == UNDEFINED_UINT )
    {
      r = FALSE;
      goto end;
    }

    if ( d0 != UNDEFINED_UINT && d1 != UNDEFINED_UINT )
    {
      if ( dict[face].dot[d0] == UNDEFINED_UINT )
      {
        dict[face].dot[d0] = d1;
      }
      else
      {
        if (dict[face].dot[d0] != d1)
        {
          r = FALSE;
          goto end;
        }
      }
    }
  }
  dict[face].done = TRUE;

  end:
  return r;
}


void
Graph::
space( /*std*/::ostream &o, uint n )
{
  uint i;
  for ( i = 0; i < n; i++ ) { o << " "; }
}


/*
  ----------------------------------------------------------------
  GraphVector
  ----------------------------------------------------------------
*/


FaceX const *GraphVector::face_list_face = (FaceX *)NULL;

void
GraphVector::
compute() throw ( Exception )
{
  Graph graph(_face);
  construct(graph);
  recurse( graph, 0 );
  uniquify();
}

void
GraphVector::
construct( Graph &graph )
{
  /*
    Construct all connection graphs for a given curve in R^2.

    ALGORITHM:
    The algorithm is a smart recursive exhaustive search. Given a partially
    completed graph, the algorithm does the following:

    STEP 1: Search in the graph for a (face, curve-edge, graph-dot) triple
      which does not have a graph edge but needs one. If there are none of
      these, the graph is complete.

    STEP 2: Separately add each possible different graph-edge to the
       (face, curve-edge, graph-dot) triple chosen in step 1 and recurse.
  */

  // initialize a new graph
  /*std*/::vector<FaceElement>::const_iterator i;
  for ( i = _face.face().begin(); i != _face.face().end(); i++ )
  {
    // initialize dots
    Graph::Dot v;
    fill( v, UNDEFINED_UINT, (*i).edge.size());
    Graph::FaceDot dot;
    if ( (*i).degree == UNDEFINED_INT )
    {
      throw ERR_INTERNAL;
    }

    if ( (*i).degree < 0 )
    {
      throw ERR_INTERNAL;
    }

    fill( dot, v, (*i).degree);
    graph._graph.push_back(dot);
  }

#if 0
  dump(/*std*/::cerr);
#endif
  // compute counts
  {
    // allocate _connection_count
    graph._connection_count.reserve(_face.size());

    /*std*/::vector<FaceElement>::const_iterator i;
    for ( i = _face.face().begin(); i != _face.face().end(); i++ )
    {
      vector<uint> v;
      fill( v, UNDEFINED_UINT, (*i).edge.size() );
      graph._connection_count.push_back(v);
    }

    // assign _connection_count[face][edge]=n, wnere n is the
    // number of graph connections that must be made from face across edge
    for ( i = _face.face().begin(); i != _face.face().end(); i++ )
    {
      uint face = i - _face.face().begin();
      uint d0 = (uint)(*i).degree;
      // cycle through the edges of the face
      /*std*/::vector<FaceElement::EdgeOfFace>::const_iterator j;
      for ( j = (*i).edge.begin();
            j != (*i).edge.end(); j++ )
      {
        uint f = (*j).opp_face;
        uint d1 = (uint)_face[f].degree;
        uint edge = j - (*i).edge.begin();
        graph._connection_count[face][edge] = Math::min(d0, d1);
      }
    }
  }

  /*
    compute _sorted_face_list
    _sorted_face_list is a list of face-edge pairs (f, e) such that
      (1) f has degree n, and the adjacent face across e has degree n+1
      (2) pairs are sorted in order of increasing degree of f
    e is the edge index, not the edge itself.
    _sorted_face_list is used by choose_connection to select new vertices
    for the connection graph.
  */

  {
  // cycle through the faces
  /*std*/::vector<FaceElement>::const_iterator i;
  for ( i = _face.face().begin(); i != _face.face().end(); i++ )
  {
    if ( (*i).degree != 0 )
    {
      // cycle through the edges of the face
      /*std*/::vector<FaceElement::EdgeOfFace>::const_iterator j;
      for ( j = (*i).edge.begin();
            j != (*i).edge.end(); j++ )
      {
        // store if degree of current face is less than degree of
        // adjacent face across edge
        if ( (*i).degree < _face.face()[(*j).opp_face].degree )
        {
          _sorted_face_list.push_back(
            /*std*/::pair<uint,uint>(
              i - _face.face().begin(), j - (*i).edge.begin() ) );
        }
    }
  }
  }
  // sort
  face_list_face = &_face;
  /*std*/::sort( _sorted_face_list.begin(), _sorted_face_list.end(),
        face_list_compare );
  }
}

bool
GraphVector::
face_list_compare(
  /*std*/::pair<uint, uint> const &a, /*std*/::pair<uint, uint> const &b )
{
  return (*face_list_face)[a.first].degree <
    (*face_list_face)[b.first].degree;
}


void
GraphVector::
recurse(
  Graph &graph,
  uint level )
{

  Graph::FaceDotEdge x, y;

  // pick out a face/edge/dot that has to be connected
  graph.choose_connection( x, y, _sorted_face_list );

  // check if there are any more connections to be made
  if ( x.face == UNDEFINED_UINT )
  {
    if ( graph.diamond_check() )
    {
      _graph.push_back(graph);
#ifdef SEARCH_TRACE
      Graph::space( /*std*/::cerr, level );
      /*std*/::cerr << "[" << level << "] ";
      /*std*/::cerr << "SUCCESS\n";
#endif
    }
    else
    {
#ifdef SEARCH_TRACE
      Graph::space( /*std*/::cerr, level );
      /*std*/::cerr << "[" << level << "] ";
      /*std*/::cerr << "FAIL\n";
#endif
    }
    return;
  }

//  /*std*/::cerr << "Running through face=" << opp_face << /*std*/::endl;
  while ( graph.next_connection( y ) )
  {
#ifdef SEARCH_TRACE
    Graph::space( /*std*/::cerr, level );

    /*std*/::cerr << "[" << level << "] ";
    /*std*/::cerr << x << " " << y << "\n";
#endif
    
    if ( ! graph.diamond_check( x, y ) ) { continue; }

    // Make a local copy
//    Graph graph1 = graph;

    // make the connection
    graph.insert_edge( x, y );

    // Recurse.
    recurse( graph, level+1 );

    graph.remove_edge( x, y );
  }
}

void
GraphVector::
uniquify()
{
  uint i;
  // sort graphs into equivalence classes

  // L holds the unsorted graphs
  /*std*/::list<uint> L;
  for ( i = 0;  i < _graph.size(); i++ )
  {
    L.push_back(i);
  }

  // G holds one representative from each equivalence class
  /*std*/::list<uint> G;

  while (!L.empty())
  {
    // put the first element of L into G
    uint g = L.front();
    L.pop_front();
    G.push_back(g);

    // remove elements in L that are equivalent to this element
    /*std*/::list<uint>::iterator q = L.begin();
    for(;;)
    {
      if (q == L.end()) { break; }
      /*std*/::list<uint>::iterator q1 = q;
      q++;
#if 0
      /*std*/::cerr << "checking graphs " << g << " and " << *q1 << /*std*/::endl;
#endif
      if ( _graph[g].same( _graph[*q1] ) )
      {
        L.erase(q1);
      }
    }
  }
  /*std*/::vector<Graph> graph;
  /*std*/::list<uint>::const_iterator q;
  for ( q = G.begin(); q != G.end(); q++ )
  {
    graph.push_back( _graph[*q] );
  }
  _graph = graph;
}

void
GraphVector::
dump( /*std*/::ostream &o )
{
  /*std*/::vector<Graph>::const_iterator i;

  uint c = 0;
  for ( i = _graph.begin(); i != _graph.end(); i++ )
  {
    o << "GRAPH " << c++ << ":\n";
    (*i).dump( o );
    o << /*std*/::endl;
  }
}

/*std*/::ostream &
operator<<( /*std::*/::ostream &o, Graph::FaceDotEdge const &x )
{
 o << "(" << (int)x.face << " " << (int)x.edge << " " << (int)x.dot << ")";
 return o;
}

} // namespace curve
